%% DEMOPAIR   A demonstration of CPair and Double-Double Arithmetic

%% Pair arithmetic
% Pair arithmetic increases the precision by storing pairs of numbers, the
% sum of which is regarded as the result. The idea goes at least back to
% Dekker's 1971 paper "A Floating-Point Technique for Extending the
% Available Precision" in Numerische Mathematik. 
%
% A key ingredient are so-called error-free transformations. For example,
% for two floating-point numbers a,b a result x,y can be computed such that
% a+b=x+y or a*b=x+y such that x is equal to the rounded-to-nearest result.
% 
% There are two toolboxes in INTLAB for a pair arithmetic, the well-known
% double-double arithmetic "dd" and a new "cpair" based on
%
%  M. Lange and S.M. Rump. Faithfully Rounded Floating-point Computations. 
%     to appear in ACM TOMS, 2018.
%
% A simple example is as follows:

format short
e = 1e-25;
x = ( 1 + e ) - 1
E1 = cpair(e);  X1 = ( 1 + E1 ) - 1
E2 = dd(e);     X2 = ( 1 + E2 ) - 1

%%
% The value of e is too small to be captured by ordinary binary64
% floating-point arithmetic; a pair arithmetic yields the correct answer.

%% A typical cancellation
% Given an approximation xs of the solution of a linear system Ax = b, the
% backward error can be improved by a residual iteration in working
% precision. To improve the forward error, the residual has to be computed in
% some higher precision. 
%
% Following we generate the inverse Hilbert matrix of dimension 9. 

n = 9;
[A,c] = hilbert(n)

%%
% We generate a right hand side by b = A*ones(n,1). 
% Note that the matrix entries are integers, so that the true solution of Ax = b is 
% indeed the vector of 1's. 
% The condition number is about 5e11, so by a
% well-known rule of thumb an approximation is expected to have about 5
% correct digits. 

format long
b = A*ones(n,1);
xs = A \ b
error_xs = norm(xs-1)

%%
% A residual in working precision cannot improve the forward error:

xs1 = xs + A \ (b - A*xs)
error_xs1 = norm(xs1-1)

%%
% That changes when computing the residual using a pair arithmetic and
% rounding the result to working precision:

xs2 = xs + A \ double(b - A*cpair(xs))
error_xs2 = norm(xs2-1)

%%
% The quadratic convergence of the Newton iteration roughly doubles the
% number of correct digits with each iteration:

xs3 = xs2 + A \ double(b - A*cpair(xs2))
error_xs3 = norm(xs3-1)

%% A famous example
% I constructed the following example to show that computation in IEEE754
% single, double and extended format may yield the same but entirely
% incorrect result. That example was analyzed several times in the
% literature:

a = 77617;
b = 33096;
f = @(a,b) 21*b^2 - 2*a^2 + 55*b^4 - 10*a^2*b^2 + a/(2*b)
y = f(a,b)

%%
% The floating-point approximation y, here computed in double precision, is
% positive. Computation in IEEE754 single or extended precision delivers the same
% result. However, the true result is negative and correctly computed in
% pair-arithmetic:

ycpair = f(cpair(a),cpair(b))
ydd = f(dd(a),dd(b))

%% A famous example II
% The next example I constructed in 1983 to show that on IBM S/370 
% architectures, the result of an expression may also be the same in
% single, double and extended precision, but wrong. The values for a and b
% are as before.

g = @(a,b) 333.75*b^6 + a^2*(11*a^2*b^2-b^6-121*b^4-2) + 5.5*b^8 + a/(2*b)
z = g(a,b)
zcpair = g(cpair(a),cpair(b))
zdd = g(dd(a),dd(b))

%%
% The floating-point approximation z is large and wrong.
% One might conclude that pair-arithmetic computes a correct approximation.
% That is not true. The
% true result is -0.8273960599468213, rounded to 16 decimal places. That is
% the same as for the previous example. 
%
% Pair-arithmetic is no panacea, it computes an approximation to the true
% result as long as the problem is not too ill-conditioned. For the first
% example the condition number is about 1e16, for the second example it
% is larger than 8e32, even too much for pair-arithmetic.

%% Cpair and dd arithmetic
% The result is the same when using the dd arithmetic. We may compare the
% accuracy and timing for a larger example. 
% We first generate a random matrix of dimension 500 with 
% approximate condition number 1e12 and calculate an approximate inverse R.

format short
n = 500;
A = randmat(n,1e12);
R = inv(A);

%%
% It can be expected that the residual I-R*A is of the order 1e-4. 
% The result and timing for cpair and dd arithmetic is as follows:

tic
res_cpair = double(eye(n) - R*cpair(A));
t_cpair = toc
norm_res_cpair = norm(res_cpair)

tic
res_dd = double(eye(n) - R*dd(A));
t_dd = toc
norm_res_dd = norm(res_dd)

%%
% The norm of the residual is the same in both cases, but cpair is faster.
% That is because final normalizations in the operations are omitted,
% often without much influencing the accuracy of the results.

%% Inversion of very ill-conditioned matrices
% In the 1980's I developed a method to calculate an approximate inverse of
% a very ill-conditioned matrix, see
%
%  S.M. Rump. Approximate inverses of almost singular matrices still contain 
%    useful information. Technical Report 90.1, TU Hamburg, 1990.
%
% Later I analyzed the method in
%
%  S.M. Rump: Inversion of extremely ill-conditioned matrices in
%  floating-point. Japan J. Indust. Appl. Math. (JJIAM), 26:249-277, 2009.
%
% A given approximate inverse R calculated in working precision is replaced
% by inv(R*A)*R, where the product R*A is calculated in double the working
% precision and rounded back into working precision, whereas the final
% product is stored in two parts:

RA = double(R*cpair(A));
invRA = inv(RA);
RR = invRA*cpair(R);
res_RR = double(eye(n) - RR*A);
norm_res_RR = norm(res_RR)

%%
% Note that the product RR*A is computed using our pair arithmetic. Now the
% norm of the residual is close to relative rounding unit of double
% precision.

%% The Wilkinson polynomial
% The famous Wilkinson polynomial has degree 20 with roots 1,2,...,20. It
% is known to be very ill-conditioned:

P = poly(cpair(1:20))

%%
% The exact coefficients are not representable in double precision.
% Evaluating the polynomial at its roots yields exactly zero:

y = polyval(P,1:20)

%%
% Using pair arithmetic one can see the large condition from the plot of the polynomial:

close all
x = linspace(1,20,1000);
plot(x,double(polyval(P,x)),x,0*x)

%% Polynomial evaluation
% Consider the following polynomial:

p = [ 4651245 -2080100 -930249 416020]

%%
% We may ask whether that polynomial has positive roots. The following plot
% uses standard double precision:

x = linspace(0.447213588,0.44721360,1000);
close
plot(x,polyval(p,x),x,0*x)

%%
% From the plot it is not clear whether there are two roots, a double one
% or no one. With the pair arithmetic we can zoom much further into the critical
% range:

x = linspace(0.4472135954995,0.4472135955002,1000);
plot(x,double(polyval(p,cpair(x))),x,0*x)

%%
% Using the pair arithmetic the polynomial values near the roots are
% calculated accurately. Thus, the roots may be found by a Newton
% iteration:

ps = [ 13953735 -4160200 -930249]

format long
K = 27;
x = cpair(0.44721);
for i=1:K
  x = x - polyval(p,x)/polyval(ps,x);
end
x1 = x

x = cpair(0.44722);
for i=1:K
  x = x - polyval(p,x)/polyval(ps,x);
end
x2 = x

%%
% Since there is an almost-double root, the convergence is slow: The
% approximations x1 and x2 coincide in 12 decimal figures. 
%
% From the
% function values at x1, x2 and at the midpoint, existence of the two roots
% is shown - at least numerically:

xm = ( x1 + x2 ) / 2
y1 = polyval(p,x1)
ym = polyval(p,xm)
y2 = polyval(p,x2)

%% Polynomial evaluation II
% Now consider the following polynomial:

p = [ 194481     -629748      993178     -782664      233290 ]

%%
% We may ask again whether this polynomial has positive roots. The following plot
% uses standard precision:

m = 0.80962343;
e = 4e-8; 
x = linspace(m-e,m+e,1000);
close
plot(x,polyval(p,x),x,0*x)

%%
% From the plot again it is not clear whether there are two roots, a double one
% or no one. With the pair arithmetic we calculate the minimum value over
% the critical range:

y = double(polyval(p,cpair(x)));
close
plot(x,y,x,0*x)

min_y = min(y)

%%
% Now there is numerical evidence that in this case the polynomial has no real root
% near m. The minimum value of 2e-11 is too small to be calculated accurately in double precision.
% Thus the previous picture showed nothing but cancellation errors.

%% Polynomial evaluation III
% As has been said, also pair-arithmetic has its limits. Once the condition
% number exceeds about 1e30 or so, we cannot expect correct answers.
% The following polynomial was constructed by Paul Zimmermann. 

p = [2280476932 84854892153 114714036446 39504798003]

%%
% Again the question is whether there are real roots near -0.6954. The
% following graph is produced using double-double arithmetic; the graph is
% the same for cpair.

xs = -0.6954389494128404;
e = 3e-16;
E = linspace(-e,e,1000);
y = double(polyval(p,dd(xs)+E));
close
plot(E,y,E,0*E)

%%
% Computing in even higher precision reveals that there is no real root
% near xs. Note that the minimum function value is about 5e-22.

%% Least squares interpolation
% For the previous plot we may calculate a quadratic interpolation
% polynomial:

p = [2280476932 84854892153 114714036446 39504798003]
xs = -0.6954389494128404;
e = 3e-16;
E = linspace(-e,e,1000)';
y = double(polyval(p,dd(xs)+E));

A = [E.^2 E ones(size(E))];
quadpoly = double(A)\y

%%
% However, due to the ill-conditioning of the problem the result is just a
% constant. Using pair-arithmetic we may compute the true interpolation
% polynomial and display it.

format long e
[Q,R] = house_qr(A);
b = Q'*y;
QuadPoly = backward(R(1:3,:),b(1:3))

close, hold on
plot(E,double(polyval(QuadPoly,E)),'r','LineWidth',1.5)
hold off

%% Enjoy INTLAB
% INTLAB was designed and written by S.M. Rump, head of the Institute for Reliable Computing,
% Hamburg University of Technology. Suggestions are always welcome to rump (at) tuhh.de
 